/* Copyright (c) 1995-1999 NEC USA, Inc.  All rights reserved.               */
/*                                                                           */
/* The redistribution, use and modification in source or binary forms of     */
/* this software is subject to the conditions set forth in the copyright     */
/* document ("Copyright") included with this distribution.                   */

/*
 * $Id: tracer.h,v 1.7.4.1 1999/02/03 22:35:51 steve Exp $
 */

/* This file contains all the prototypes & globals used in handling ping and */
/* tracerotue                                                                */
#ifndef TRACER_H
#define TRACER_H

extern int  PTSetup   P((S5IOInfo *, S5LinkInfo *, S5CommandInfo *));

#endif

