/* Copyright (c) 1995-1999 NEC USA, Inc.  All rights reserved.               */
/*                                                                           */
/* The redistribution, use and modification in source or binary forms of     */
/* this software is subject to the conditions set forth in the copyright     */
/* document ("Copyright") included with this distribution.                   */

/*
 * $Id: packet.h,v 1.4.4.1 1999/02/03 22:35:38 steve Exp $
 */

#ifndef PACKET_H
#define PACKET_H

extern int PacketPrintSetup P((S5LinkInfo *, S5FilterInfo *));

#endif
