/*                                                                           */
/*  Copyright (c) 1992, 1993                                                 */
/*   The Regents of the University of California.  All rights reserved.      */
/*                                                                           */
/*   Redistribution and use in source and binary forms, with or without      */
/*   modification, are permitted provided that the following conditions      */
/*   are met:                                                                */
/*   1. Redistributions of source code must retain the above copyright       */
/*   notice, this list of conditions and the following disclaimer.           */
/*   2. Redistributions in binary form must reproduce the above copyright    */
/*   notice, this list of conditions and the following disclaimer in the     */
/*   documentation and/or other materials provided with the distribution.    */
/*   3. All advertising materials mentioning features or use of this         */
/*   software must display the following acknowledgement:                    */
/*                                                                           */
/*     This product includes software developed by the University of         */
/*     California, Berkeley and its contributors.                            */
/*                                                                           */
/*   4. Neither the name of the University nor the names of its contributors */
/*   may be used to endorse or promote products derived from this software   */
/*   without specific prior written permission.                              */
/*                                                                           */
/*   THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND */
/*   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   */
/*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR      */
/*   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS  */
/*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR  */
/*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF    */
/*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR         */
/*   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   */
/*   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR */
/*   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF  */
/*   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                              */

#ifdef SHOW_SCCSIDS
static char sccsid[] = "@(#)read_password.c	8.3 (Berkeley) 5/30/95";
#endif

/*  Copyright 1985, 1986, 1987, 1988 by the Massachusetts Institute          */
/*  of Technology.                                                           */
/*                                                                           */
/*   For copying and distribution information, please see the file           */
/*   <mit-copyright.h>.                                                      */
/*                                                                           */
/*   This routine prints the supplied string to standard                     */
/*   output as a prompt, and reads a password string without                 */
/*   echoing.                                                                */

