/* Copyright (c) 1995-1999 NEC USA, Inc.  All rights reserved.               */
/*                                                                           */
/* The redistribution, use and modification in source or binary forms of     */
/* this software is subject to the conditions set forth in the copyright     */
/* document ("Copyright") included with this distribution.                   */

/*
 * $Id: proxy.h,v 1.11.4.1 1999/02/03 22:35:40 steve Exp $
 */

#ifndef PROXY_H
#define PROXY_H

extern int HandlePxyConnection P((S5IOHandle));

#endif
